WeaponUnderbarrelRaycast = WeaponUnderbarrelRaycast or class(WeaponUnderbarrel)
WeaponUnderbarrelRaycast.GADGET_TYPE = "underbarrel_launcher"

local mvec_to = Vector3()
local mvec_spread_direction = Vector3()

function WeaponUnderbarrelRaycast:_fire_raycast(weapon_base, user_unit, from_pos, direction, dmg_mul, shoot_player, spread_mul, autohit_mul, suppr_mul, shoot_through_data)
	local result = {}
	local spread_x, spread_y = weapon_base:_get_spread(user_unit)
	local ray_distance = weapon_base:weapon_range()
	local right = direction:cross(Vector3(0, 0, 1)):normalized()
	local up = direction:cross(right):normalized()
	local theta = math.random() * 360
	local ax = math.sin(theta) * math.random() * spread_x * (spread_mul or 1)
	local ay = math.cos(theta) * math.random() * spread_y * (spread_mul or 1)

	mvector3.set(mvec_spread_direction, direction)
	mvector3.add(mvec_spread_direction, right * math.rad(ax))
	mvector3.add(mvec_spread_direction, up * math.rad(ay))
	mvector3.set(mvec_to, mvec_spread_direction)
	mvector3.multiply(mvec_to, ray_distance)
	mvector3.add(mvec_to, from_pos)

	local damage = weapon_base:_get_current_damage(dmg_mul)
	local ray_hits, hit_enemy = weapon_base:_collect_hits(from_pos, mvec_to)
	local hit_anyone = false
	local auto_hit_candidate, suppression_enemies = weapon_base:check_autoaim(from_pos, direction)

	if suppression_enemies and weapon_base._suppression then
		result.enemies_in_cone = suppression_enemies
	end

	if weapon_base._autoaim then
		local weight = 0.1

		if auto_hit_candidate and not hit_enemy then
			local autohit_chance = 1 - math.clamp((weapon_base._autohit_current - weapon_base._autohit_data.MIN_RATIO) / (weapon_base._autohit_data.MAX_RATIO - weapon_base._autohit_data.MIN_RATIO), 0, 1)

			if autohit_mul then
				autohit_chance = autohit_chance * autohit_mul
			end

			if math.random() < autohit_chance then
				weapon_base._autohit_current = (weapon_base._autohit_current + weight) / (1 + weight)

				mvector3.set(mvec_to, from_pos)
				mvector3.add_scaled(mvec_to, auto_hit_candidate.ray, ray_distance)

				ray_hits, hit_enemy = weapon_base:_collect_hits(from_pos, mvec_to)
			end
		end

		if hit_enemy then
			weapon_base._autohit_current = (weapon_base._autohit_current + weight) / (1 + weight)
		elseif auto_hit_candidate then
			weapon_base._autohit_current = weapon_base._autohit_current / (1 + weight)
		end
	end

	local hit_count = 0
	local cop_kill_count = 0
	local hit_through_wall = false
	local hit_through_shield = false
	local hit_result = nil

	for _, hit in ipairs(ray_hits) do
		damage = weapon_base:get_damage_falloff(damage, hit, user_unit)
		hit_result = weapon_base._bullet_class:on_collision(hit, weapon_base._unit, user_unit, damage)

		if hit_result and hit_result.type == "death" then
			local unit_type = hit.unit:base() and hit.unit:base()._tweak_table
			local is_civilian = unit_type and CopDamage.is_civilian(unit_type)

			if not is_civilian then
				cop_kill_count = cop_kill_count + 1
			end
		end

		if hit_result then
			hit.damage_result = hit_result
			hit_anyone = true
			hit_count = hit_count + 1
		end

		if hit.unit:in_slot(managers.slot:get_mask("world_geometry")) then
			hit_through_wall = true
		elseif hit.unit:in_slot(managers.slot:get_mask("enemy_shield_check")) then
			hit_through_shield = hit_through_shield or alive(hit.unit:parent())
		end

		if hit_result and hit_result.type == "death" and cop_kill_count > 0 then
			local unit_type = hit.unit:base() and hit.unit:base()._tweak_table
			local multi_kill, enemy_pass, obstacle_pass, weapon_pass, weapons_pass, weapon_type_pass = nil
		end
	end

	result.hit_enemy = hit_anyone

	if weapon_base._autoaim then
		weapon_base._shot_fired_stats_table.hit = hit_anyone
		weapon_base._shot_fired_stats_table.hit_count = hit_count

		if (not weapon_base._ammo_data or not weapon_base._ammo_data.ignore_statistic) and not weapon_base._rays then
			managers.statistics:shot_fired(weapon_base._shot_fired_stats_table)
		end
	end

	local furthest_hit = ray_hits[#ray_hits]

	if (furthest_hit and furthest_hit.distance > 600 or not furthest_hit) and alive(weapon_base._obj_fire) then
		weapon_base._obj_fire:m_position(weapon_base._trail_effect_table.position)
		mvector3.set(weapon_base._trail_effect_table.normal, mvec_spread_direction)

		local trail = World:effect_manager():spawn(weapon_base._trail_effect_table)

		if furthest_hit then
			World:effect_manager():set_remaining_lifetime(trail, math.clamp((furthest_hit.distance - 600) / 10000, 0, furthest_hit.distance))
		end
	end

	if weapon_base._alert_events then
		result.rays = ray_hits
	end

	return result
end